/**
 * @file lookup_helpers.c common helper functions used by various examples.
 *
 * @author Matthew Tesch < matt @ hebirobotics.com >
 * @since 29 May 2015
 */

#include "hebi_lookup.h"
#include "hebi_group.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/**
 * Check whether this argument list has a 'verbose' flag. Returns '1' for yes,
 * '0' for no.
 *
 * @author Matthew Tesch < matt @ hebirobotics.com >
 * @since 23 Sep 2015
 */ 
int hasVerboseFlag(int argc, char* argv[])
{
  int i;
  for (i = 1; i < argc; i++)
  {
    if (strncmp(argv[i], "-v", 2) == 0)
      return 1;
  }
  return 0;
}
/**
 * Check whether this argument list has a 'help' flag. Returns '1' for yes,
 * '0' for no.
 *
 * @author Matthew Tesch < matt @ hebirobotics.com >
 * @since 23 Sep 2015
 */ 
int hasHelpFlag(int argc, char* argv[])
{
  int i;
  for (i = 1; i < argc; i++)
  {
    if (strncmp(argv[i], "-h", 2) == 0)
      return 1;
  }
  return 0;
}

/**
 * @param printModuleList set to '1' to print the module list during lookup.
 */
HebiGroupPtr getConnectedGroupFromName(const char* family_name, const char* name, int printModuleList)
{
  // Give the modules plenty of time to appear.
  long timeout_ms = 4000;

  // Create the lookup.
  HebiLookupPtr lookup = hebiLookupCreate();
  if (lookup == NULL)
    return NULL;

  // Create the group.
  // TODO: change the arg order here!
  HebiGroupPtr group = hebiCreateConnectedGroupFromName(lookup, name, family_name, timeout_ms);

  if (printModuleList == 1)
    hebiPrintLookupTable(lookup);

  // Clean up the lookup no matter whether finding group was successful or not.
  hebiLookupRelease(lookup);
  return group;
}

/**
 * @param printModuleList set to '1' to print the module list during lookup.
 */
HebiGroupPtr getGroupFromNames(const char* const* family_names, int num_family_names,
                             const char* const* names, int num_names, int printModuleList)
{
  // Give the modules plenty of time to appear.
  long timeout_ms = 4000;

  // Create the lookup.
  HebiLookupPtr lookup = hebiLookupCreate();
  if (lookup == NULL)
    return NULL;

  // Create the group.
  // TODO: change the arg order here!
  HebiGroupPtr group = hebiCreateGroupFromNames(lookup, names, num_names, family_names, num_family_names, timeout_ms);

  if (printModuleList == 1)
    hebiPrintLookupTable(lookup);

  // Clean up the lookup no matter whether finding group was successful or not.
  hebiLookupRelease(lookup);
  return group;
}

/*
 * 1 if yes, 0 if no.
 */
int isHexDigit(char c)
{
  return ((c >= '0' && c <= '9') ||
          (c >= 'a' && c <= 'f') ||
          (c >= 'A' && c <= 'F'));
}

/**
 * Is this a valid string of format dd:dd:dd:dd:dd:dd, where 'd' is a hex digit
 * 0-F.  Lowercase values accepted. 1 if yes, 0 if no.
 */
int isHexString(char* str)
{
  int len = strlen(str);
  if (len != 17) // 6 * 2 (digits) + 5 (":"s)
    return 0;
  int j;
  for (j = 2; j < len; j += 3)
    if (str[j] != ':')
      return 0;
  for (j = 0; j < len; j += 3)
    if (!isHexDigit(str[j]))
      return 0;
  for (j = 1; j < len; j += 3)
    if (!isHexDigit(str[j]))
      return 0;
  return 1;
}

uint8_t hexToInt(char c)
{
  if (c >= '0' && c <= '9')
    return c - '0';
  if (c >= 'a' && c <= 'f')
    return c - 'a' + 10;
  if (c >= 'A' && c <= 'F')
    return c - 'A' + 10;
  return 0;
}

uint8_t byteFromHexPair(char c1, char c2)
{
  return hexToInt(c1) * 16 + hexToInt(c2);
}

/**
 * @param printModuleList set to '1' to print the module list during lookup.
 */
HebiGroupPtr getConnectedGroupFromMac(const char* h1, const char* h2,
                                    const char* h3, const char* h4,
                                    const char* h5, const char* h6, int printModuleList)
{
  // Parse the mac address.
  HebiMacAddress mac;
  mac.bytes_[0] = byteFromHexPair(h1[0], h1[1]);
  mac.bytes_[1] = byteFromHexPair(h2[0], h2[1]);
  mac.bytes_[2] = byteFromHexPair(h3[0], h3[1]);
  mac.bytes_[3] = byteFromHexPair(h4[0], h4[1]);
  mac.bytes_[4] = byteFromHexPair(h5[0], h5[1]);
  mac.bytes_[5] = byteFromHexPair(h6[0], h6[1]);

  // Give the modules plenty of time to appear.
  long timeout_ms = 4000;

  // Create the lookup.
  HebiLookupPtr lookup = hebiLookupCreate();
  if (lookup == NULL)
    return NULL;

  // Create the group.
  HebiGroupPtr group = hebiCreateConnectedGroupFromMac(lookup, &mac, timeout_ms);

  if (printModuleList == 1)
    hebiPrintLookupTable(lookup);

  // Clean up the lookup no matter whether finding group was successful or not.
  hebiLookupRelease(lookup);
  return group;
}

/**
 * @param printModuleList set to '1' to print the module list during lookup.
 */
HebiGroupPtr getGroupFromMacs(const char* const* mac_list, int num_macs, int printModuleList)
{
  // Parse the mac address.
  HebiMacAddress* macs = (HebiMacAddress*)calloc(num_macs, sizeof(HebiMacAddress));
  int i;
  for (i = 0; i < num_macs; i++)
  {
    macs[i].bytes_[0] = byteFromHexPair(mac_list[i][0], mac_list[i][1]);
    macs[i].bytes_[1] = byteFromHexPair(mac_list[i][3], mac_list[i][4]);
    macs[i].bytes_[2] = byteFromHexPair(mac_list[i][6], mac_list[i][7]);
    macs[i].bytes_[3] = byteFromHexPair(mac_list[i][9], mac_list[i][10]);
    macs[i].bytes_[4] = byteFromHexPair(mac_list[i][12], mac_list[i][13]);
    macs[i].bytes_[5] = byteFromHexPair(mac_list[i][15], mac_list[i][16]);
  }

  // Give the modules plenty of time to appear.
  long timeout_ms = 4000;

  // Create the lookup.
  HebiLookupPtr lookup = hebiLookupCreate();
  if (lookup == NULL)
    return NULL;

  // Create the group.
  HebiGroupPtr group = hebiCreateGroupFromMacs(lookup, macs, num_macs, timeout_ms);

  if (printModuleList == 1)
    hebiPrintLookupTable(lookup);

  // Clean up the lookup no matter whether finding group was successful or not.
  hebiLookupRelease(lookup);
  return group;
}

/**
 * Returns 'null' if an error occurs, otherwise a malloc-ed array of pointers
 * to the start of null-terminated substrings. The ref. variable 'len' is set
 * to the number of returned substrings.
 *
 * 'mutable_src' is modified; ',' tokens are replaced with '\0' null characters.
 */
char** parseCommaDelimitedArgs(char* mutable_src, int* len)
{
  (*len) = 1;
  int src_len = strlen(mutable_src);
  // Count the number of commas:
  int i;
  for (i = 0; i < src_len; i++)
  {
    if (mutable_src[i] == ',')
      (*len)++;
  }
  // Add the location after each comma to the list of strings.
  char** res = (char**)calloc((*len), sizeof(char*));
  res[0] = mutable_src;
  int res_idx = 1;
  for (i = 1; i < src_len; i++)
  {
    if (mutable_src[i] == ',')
    {
      mutable_src[i] = '\0';
      res[res_idx] = mutable_src + i + 1;
      res_idx++;
    }
  }
  return res;
}

/**
 * Allocates space for and copies the source string, saving its length to 'len'.
 */
char* mutableStrCopy(const char* str_src)
{
  int len = strlen(str_src);
  char * pair_src = (char*)malloc(len + 1);
  strncpy(pair_src, str_src, len);
  pair_src[len] = '\0';
  return pair_src;
}

/**
 * Parse the given arguments, and look up group based on the given arguments.
 *
 * @author Matthew Tesch < matt @ hebirobotics.com >
 * @since 29 May 2015
 */
HebiGroupPtr getGroupFromArgs(int argc, char* argv[])
{
  // Display help and exit (if help arg or no args)
  int hasHelp = hasHelpFlag(argc, argv);
  if (argc == 1 || hasHelp)
  {
    if (argc == 1)
      printf("\nNo arguments given -- usage instructions given below:\n");
    printf("\n\
 Arguments are parsed to determine search terms for groups.\n\
 User can either provide\n\
  1) single module specification to form connected group (-p, -n/-f, or -m)\n\
  2) list of families and list of names independently (-F and -N)\n\
  3) list of family/name pairs (-P)\n\
  4) list of mac addresses (-M)\n\
 Whichever option is found first is used; all others are ignored.\n\
\n\
 Argument notes:\n\
 '-p', '-n', '-f', and '-m' are as for single modules\n\
 '-N' argument is list of names (use with '-F'), and should be followed by:\n\
  'name_1,name_2,...,name_n'\n\
 '-F' argument is list of families (use with '-N'), and should be followed by:\n\
  'family_1,family_2,...,family_n'\n\
 '-P' argument is list of family name/name pairs, and should be followed by:\n\
  'pair_1,pair_2,...,pair_n' where each pair_i is 'family|name'\n\
 '-M' argument is list of macs, and should be followed by:\n\
  'm_1,m_2,...,m_n' where each m_i is '01:23:45:ab:cd:ef'\n\
\n\
 examples:\n\
 ./script -nModule05 -fArm\n\
 ./script -pArm|Module05\n\
 ./script -m01:23:45:ab:cd:ef\n\
 ./script -NModule05,Module07,Module02 -FArm,Arm,Arm\n\
 ./script -NModule05,Module07,Module02 -FArm\n\
 ./script -PArm|Module05,Arm|Module06\n\
 ./script -M01:23:45:ab:cd:ef,05:34:25:a4:fe:3d\n\
\n\
 Incorrectly formatted arguments are ignored. Only first full identifier set\n\
 is used (e.g., a '-N', '-P', '-F' set of args will only use the info in the\n\
 '-P', while a '-N', '-F', '-P' set will only use the info in '-N' and '-F'\n\
 together).\n\
\n\
 When '-v' is used, verbose output is enabled to help diagnose lookup issues.\n\
\n\
 '-h' prints this help and exits.\n\n");
    return NULL;
  }
  HebiGroupPtr group = NULL;
  // Variables to track data passed in if both a '-n' and '-f' argument are
  // passed in.  The first is saved, and then the second of these triggers the
  // lookup.
  int has_family = 0;
  int has_name = 0;
  // 'other' means the other of name/family; whichever is being saved.
  char** list_of_other_names = NULL;
  char* other_name_src = NULL;
  int num_of_other_names = 0;
  // Store potential single name and/or family for connected groups.
  char* single_name = NULL;
  char* single_family = NULL;

  int printModuleList = hasVerboseFlag(argc, argv);

  // Parse the arguments and return any group found.  (Start at 1 because we
  // don't want to use the program's name!)
  int i;
  for (i = 1; i < argc; i++)
  {
    if (strncmp(argv[i], "-N", 2) == 0 && !has_name)
    {
      // Just save until family is loaded
      if (!has_family)
      {
        other_name_src = mutableStrCopy(argv[i] + 2);
        list_of_other_names =
          parseCommaDelimitedArgs(other_name_src, &num_of_other_names);
        has_name = 1;
      }
      // Try and use immediately
      else
      {
        char* name_src = mutableStrCopy(argv[i] + 2);
        int num_names;
        char** list_names = parseCommaDelimitedArgs(name_src, &num_names);
        group = getGroupFromNames(
          (const char * const *)list_of_other_names, num_of_other_names,
          (const char * const *)list_names, num_names,
          printModuleList);
        free(name_src);
        free(list_names);
        break;
      }
    }
    if (strncmp(argv[i], "-F", 2) == 0 && !has_family)
    {
      // Just save until name is loaded
      if (!has_name)
      {
        other_name_src = mutableStrCopy(argv[i] + 2);
        list_of_other_names =
          parseCommaDelimitedArgs(other_name_src, &num_of_other_names);
        has_family = 1;
      }
      // Try and use immediately
      else
      {
        char* family_src = mutableStrCopy(argv[i] + 2);
        int num_families;
        char** list_families = parseCommaDelimitedArgs(family_src, &num_families);
        group = getGroupFromNames(
          (const char * const *)list_families, num_families,
          (const char * const *)list_of_other_names, num_of_other_names,
          printModuleList);
        free(family_src);
        free(list_families);
        break;
      }
    }
    else if (strncmp(argv[i], "-P", 2) == 0)
    {
      // Copy arguments so we can modify in place!
      char* pair_src = mutableStrCopy(argv[i] + 2);

      // Create arrays for family name pointers
      int num_pairs;
      char** list_families = parseCommaDelimitedArgs(pair_src, &num_pairs);

      // Split each substring to add name pointers
      int success = 1;
      char ** names = (char**)calloc(num_pairs, sizeof(char*));
      int pair_idx;
      for (pair_idx = 0; pair_idx < num_pairs; pair_idx++)
      {
        char * split_point = strpbrk(list_families[pair_idx], "|");
        if (split_point == NULL)
        {
          success = 0;
          break;
        }
        // Start the name string after the '|'
        names[pair_idx] = split_point + 1;
        // Turn the split point into a null character to term the family name
        // string.
        split_point[0] = '\0';
      }
      // Build a group if splitting pairs went well.
      if (success)
      {
        group = getGroupFromNames(
          (const char * const *)list_families, num_pairs,
          (const char * const *)names, num_pairs, printModuleList);
      }
      free(list_families);
      free(names);
      free(pair_src);
      break;
    }
    else if (strncmp(argv[i], "-M", 2) == 0)
    {
      // Copy arguments so we can modify in place!
      char* mac_src = mutableStrCopy(argv[i] + 2);

      // Create arrays for pointers to max strings
      int num_macs;
      char** list_macs = parseCommaDelimitedArgs(mac_src, &num_macs);
      int mac_idx;
      for (mac_idx = 0; mac_idx < num_macs; mac_idx++)
      {
        // Parse mac address arguments
        if (!isHexString(list_macs[mac_idx]))
        {
          printf("Invalid mac argument; format of each element in list should be 00:01:02:ab:cd:ef\n");
          continue;
        }
      }

      group = getGroupFromMacs((const char * const *)list_macs, num_macs, printModuleList);
      free(list_macs);
      free(mac_src);
      break;
    }
    if (strncmp(argv[i], "-p", 2) == 0)
    {
      // Parse name arguments (family|name)
      char * split_point = strpbrk(argv[i], "|");
      if (split_point == NULL)
      {
        printf("Invalid name argument; format should be -n<family>|<name>\n");
        continue;
      }
      // Have to copy the input name to make it null termed.  (Technically we
      // could overwrite argv[i], but this seems like a BAD idea).
      int family_length = (split_point - argv[i]) - 2;
      single_family = (char*) malloc(family_length + 1);
      strncpy(single_family, argv[i] + 2, family_length);
      single_family[family_length] = '\0';
      group = getConnectedGroupFromName(single_family, split_point + 1, printModuleList);
      break;
    }
    else if (strncmp(argv[i], "-n", 2) == 0)
    {
      // If we already have one, continue
      if (single_name != NULL)
        continue;
      // If we already have a family, try to parse now:
      if (single_family != NULL)
      {
        group = getConnectedGroupFromName(single_family, argv[i] + 2, printModuleList);
        break;
      }
      // Save name argument until we get to family
      int name_length = strlen(argv[i] + 2);
      single_name = (char*) malloc(name_length + 1);
      strcpy(single_name, argv[i] + 2);
    }
    else if (strncmp(argv[i], "-f", 2) == 0)
    {
      // If we already have one, continue
      if (single_family != NULL)
        continue;
      if (single_name != NULL)
      {
        group = getConnectedGroupFromName(argv[i] + 2, single_name, printModuleList);
        break;
      }
      // Save family argument until we get to name
      int family_length = strlen(argv[i] + 2);
      single_family = (char*) malloc(family_length + 1);
      strcpy(single_family, argv[i] + 2);
    }
    else if (strncmp(argv[i], "-m", 2) == 0)
    {
      // Parse mac address arguments
      if (!isHexString(argv[i] + 2))
      {
        printf("Invalid mac argument; format should be -m00:01:02:ab:cd:ef\n");
        continue;
      }
      group = getConnectedGroupFromMac(argv[i] + 2, argv[i] + 5,
                                       argv[i] + 8, argv[i] + 11,
                                       argv[i] + 14, argv[i] + 17,
                                       printModuleList);
      break;
    }
  }

  if (list_of_other_names != NULL)
    free(list_of_other_names);
  if (other_name_src != NULL)
    free(other_name_src);
  if (single_name != NULL)
    free(single_name);
  if (single_family != NULL)
    free(single_family);

  return group;
}
