/**
 * Demo : Torque Reaction Ping Pong
 */

#include <stdio.h>
#include <math.h>
#include <time.h>

#include "hebi_lookup.h"
#include "lookup_helpers.c"
#include "hebi_util.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

// (Initial) demo parameters
// NOTE: all are stored as globals, but could be passed in as struct as user data in callback
float desiredPosition = 0; // rad; set on initial response.
int initial_response = 1;
float desiredVelocity = -3.5; // rad/s
const float coolDownTime = 0.5; // s
const float torqueThreshold = 2; // Nm
const float arm_grav_torque = 1;
const float rate = 200; // Hz
HebiGroupPtr group = NULL;
HebiGroupCommandPtr group_command = NULL;
clock_t tic;

void feedback_callback(HebiGroupFeedbackPtr group_feedback, void * user_data)
{
  // Helper variables for later calls to high res position:
  int64_t revs;
  float offset;

  // Get convenience variables:
  HebiFeedbackPtr feedback = hebiGroupFeedbackGetModuleFeedback(group_feedback, 0);
  int has_command =
    hebiFeedbackHasFloat(feedback, FeedbackFloatTorqueCommand) &&
    hebiFeedbackHasFloat(feedback, FeedbackFloatVelocityCommand);

  if (initial_response == 1)
  {
    if (hebiFeedbackHasHighResAngle(feedback, FeedbackHighResAnglePosition))
    {
      hebiFeedbackGetHighResAngle(feedback, FeedbackHighResAnglePosition, &revs, &offset);
      desiredPosition = revs * 2.0 * M_PI + offset;
      initial_response = 0;
      // Start the timing!
      tic = clock();
    }
    else
    {
      return;
    }
  }

  // Reverse velocity direction on impact detection
  if (has_command)
  {
    float torqueErr =
      hebiFeedbackGetFloat(feedback, FeedbackFloatTorque) -
      hebiFeedbackGetFloat(feedback, FeedbackFloatTorqueCommand);
    float vel = hebiFeedbackGetFloat(feedback, FeedbackFloatVelocityCommand);
    printf("Torque_error %f\n", torqueErr);
    if (torqueErr * (fabs(vel)/vel) > torqueThreshold)
    {
      // Get cool down:
      clock_t toc = clock();
      double cpu_time_used = ((double) (toc - tic)) / CLOCKS_PER_SEC * 60;
      // Ignore triggers for the cooldown period 
      // to prevent oscillations
      printf("Big torque error %f; time %f coolDownTime %f\n", torqueErr, cpu_time_used, coolDownTime);
      if (cpu_time_used > coolDownTime)
      {
        printf("Switch\n");
        desiredVelocity = -desiredVelocity;
        tic = toc;
      }
    }
  }

  // Forward integrate velocity to get the desired position
  float dt = 1.0/rate; // OR: dt = (fbk.time - lastTime);
  desiredPosition = desiredPosition + desiredVelocity * dt;

  // Compensate for gravity by feeding forward torques
  hebiFeedbackGetHighResAngle(feedback, FeedbackHighResAnglePosition, &revs, &offset);
  float position = revs * 2.0 * M_PI + offset;
  float gravCompTorque =
    arm_grav_torque * cos(position);

  // Update joint commands
  HebiCommandPtr command = hebiGroupCommandGetModuleCommand(group_command, 0);
  hebiCommandSetHighResAngle(command, CommandHighResAnglePosition, 0, desiredPosition);
  hebiCommandSetFloat(command, CommandFloatVelocity, desiredVelocity);
  hebiCommandSetFloat(command, CommandFloatTorque, gravCompTorque);
  hebiGroupSendCommand(group, group_command);
}

int main(int argc, char* argv[])
{
  // Try and get the requested group.
  // NOTE: could hardcode module name here!
  group = getGroupFromArgs(argc, argv);
  if (group == NULL)
  {
    hebiCleanup();
    return -1;
  }
  int group_size = hebiGroupGetNumberOfModules(group);
  if (group_size != 1)
  {
    printf("Assertion that group is of size one failed!\n");
    hebiGroupRelease(group);
    hebiCleanup();
    return -1;
  }

  // Create a command object; this can be sent to the group
  group_command = hebiGroupCommandCreate(group_size);
  if (group_command == NULL)
  {
    printf("Could not create group command object!\n");
    hebiGroupRelease(group);
    hebiCleanup();
    return -1;
  }

  printf("Starting feedback thread\n");
  hebiGroupRegisterFeedbackHandler(group, feedback_callback, NULL);
  hebiGroupSetFeedbackFrequencyHz(group, 200);
  hebi_sleep_s(10);
  hebiGroupSetFeedbackFrequencyHz(group, 0);
  hebiGroupClearFeedbackHandlers(group);
  
  printf("Feedback callback thread stopped\n");

  // Stopping!

  int i; 
  for (i = 0; i < group_size; ++i)
  {
    HebiCommandPtr command = hebiGroupCommandGetModuleCommand(group_command, i);
    hebiCommandSetHighResAngle(command, CommandHighResAnglePosition, 0, NAN);
    hebiCommandSetFloat(command, CommandFloatVelocity, 0);
    hebiCommandSetFloat(command, CommandFloatTorque, 0);
  }
  printf("Got %d as hebiGroupSendCommandWithAcknowledgement response (0 indicates success)\n", 
    hebiGroupSendCommandWithAcknowledgement(group, group_command, 1000));
  hebiCleanup();
}

// TODO: ctrl-c
// TODO: timestamps for dt
